import pytest
from aiohttp import ClientSession

from restgdf.directory.directory import Directory


@pytest.mark.asyncio
async def test_directory():
    async with ClientSession() as s:
        directory = await Directory.from_url(
            "https://maps1.vcgov.org/arcgis/rest/services",
            session=s,
        )
        assert len(directory.data) > 0



"""A package for getting GeoDataFrames from ArcGIS FeatureLayers."""

from restgdf.featurelayer.featurelayer import FeatureLayer
from restgdf.directory.directory import Directory

__all__ = ["Directory", "FeatureLayer"]


__version__ = "0.6.0"



import asyncio
import re
from collections.abc import Iterable
from typing import Any

import aiohttp

ends_with_num_pat = re.compile(r"\d+$")


def ends_with_num(url: str) -> bool:
    """Return True if the given URL ends with a number."""
    return bool(ends_with_num_pat.search(url))


def where_var_in_list(var: str, vals: Iterable[str]) -> str:
    """Return a where clause for a variable in a list of values."""
    vals_str = ", ".join(f"'{val}'" for val in vals)
    return f"{var} In ({vals_str})"


async def get_services_and_folders(
    session: aiohttp.ClientSession,
    url: str,
) -> tuple[list[str], list[dict]]:
    """Asynchronously retrieve services and folders from the given ArcGIS server URL."""
    async with session.get(url, params={"f": "json"}) as resp:
        resp_dict = await resp.json()
        folders = resp_dict.get("folders", [])
        services = resp_dict.get("services", [])
        return folders, services


async def get_layers_for_service(
    session: aiohttp.ClientSession,
    service_url: str,
) -> dict[str, Any]:
    """Asynchronously retrieve layers for a single service."""
    async with session.get(service_url, params={"f": "json"}) as response:
        service_data = await response.json()
        return {
            layer["name"]: layer | {"url": f"{service_url}/{layer['id']}"}
            for layer in service_data.get("layers", [])
        }


async def fetch_all_data(
    session: aiohttp.ClientSession,
    base_url: str,
) -> dict[str, Any]:
    """Fetch all services and their layers in a highly concurrent manner."""
    # Retrieve the initial list of folders and services
    folders, services = await get_services_and_folders(session, base_url)
    # Prepare list of URLs to fetch layers from each service
    service_urls = [
        f"{base_url}/{service['name']}/{service['type']}" for service in services
    ]
    # Add nested folders' service URLs
    for folder in folders:
        nested_folders, nested_services = await get_services_and_folders(
            session,
            f"{base_url}/{folder}",
        )
        service_urls += [
            f"{base_url}/{folder}/{service['name']}/{service['type']}"
            for service in nested_services
        ]
    # Fetch all layers concurrently
    layer_tasks = [get_layers_for_service(session, url) for url in service_urls]
    results = await asyncio.gather(*layer_tasks)
    # Combine results into a single dictionary
    all_layers = {k: v for d in results for k, v in d.items()}
    return all_layers



import pytest
from aiohttp import ClientSession
from pytest import raises
from restgdf.featurelayer.featurelayer import FeatureLayer


@pytest.mark.asyncio
async def test_featurelayer():
    async with ClientSession() as s:
        # print("testing workflow")
        with pytest.raises(ValueError):
            await FeatureLayer.from_url(
                "https://maps1.vcgov.org/arcgis/rest/services/Beaches/MapServer",
                session=s,
            )
        with pytest.raises(ValueError):
            await FeatureLayer.from_url(
                "https://maps1.vcgov.org/arcgis/rest/services/Aerials/MapServer/4",
                session=s,
            )
        beachurl = r"https://maps1.vcgov.org/arcgis/rest/services/Beaches/MapServer/6"
        beaches = await FeatureLayer.from_url(
            beachurl,
            session=s,
            where="City <> 'fgsfds'",
        )
        beaches_gdf = await beaches.getgdf()
        assert len(await beaches.samplegdf(2)) == 2
        assert len(await beaches.headgdf(2)) == 2
        assert len(beaches_gdf) > 0
        assert all(
            "fgsfds" in s for s in (beaches.wherestr, beaches.kwargs["data"]["where"])
        )
        assert (
            len(await beaches.getuniquevalues(("City", "Status"), sortby="City")) > 10
        )
        daytona = await beaches.where("City LIKE 'DAYTONA%'")
        assert "Status" in daytona.fields
        assert str(beaches) == f"Beach Access Points ({beachurl})"

        zipurl = "https://services.arcgis.com/P3ePLMYs2RVChkJx/ArcGIS/rest/services/USA_ZIP_Codes_2016/FeatureServer/0"
        ziprest = await FeatureLayer.from_url(zipurl, where="STATE = 'OH'", session=s)
        testkwargs = {k: v for k, v in ziprest.kwargs.items()}
        assert "Cincinnati" in await ziprest.getuniquevalues("PO_NAME")
        assert await ziprest.getuniquevalues(
            "PO_NAME",
        ) == await ziprest.getuniquevalues(
            "PO_NAME",
        )
        assert (await ziprest.getvaluecounts("PO_NAME")).set_index("PO_NAME").to_dict()[
            "PO_NAME_count"
        ]["Cincinnati"] > 40
        with raises(IndexError):
            assert "Cincinnati" in await ziprest.getuniquevalues("zzzzzz")
        with raises(IndexError):
            assert len(await ziprest.getnestedcount(("PO_NAME", "ZIP"))) > 900
        assert len(await ziprest.getnestedcount(("PO_NAME", "ZIP_CODE"))) > 900
        assert ziprest.count > ziprest.jsondict["maxRecordCount"]
        assert len(await ziprest.getgdf()) > ziprest.jsondict["maxRecordCount"]
        assert ziprest.kwargs == testkwargs  # make sure nothing is altering kwargs



import asyncio
from unittest.mock import patch

import pytest
from aiohttp import ClientSession

from restgdf.utils import where_var_in_list
from restgdf.featurelayer._getinfo import (
    DEFAULTDICT,
    default_data,
    get_feature_count,
    get_jsondict,
    get_max_record_count,
    get_offset_range,
)

TESTJSON = {"count": 500, "maxRecordCount": 100}


def test_wherevarinlist():
    assert where_var_in_list("STATE", ["FL", "GA"]) == "STATE In ('FL', 'GA')"


# def mock_session_post(*args, **kwargs):
#     class MockResponse:
#         def __init__(self, json_data):
#             self.json_data = json_data
#
#         def json(self):
#             return self.json_data
#
#     return MockResponse(TESTJSON)


def mock_session_post(*args, **kwargs):
    class MockResponse:
        def __init__(self, json_data):
            self.json_data = json_data

        async def json(self):  # make this method async
            return self.json_data

    future = asyncio.Future()  # create a Future object
    future.set_result(MockResponse(TESTJSON))  # set the result of the future
    return future  # return the future


@pytest.mark.asyncio
@patch(
    "restgdf.featurelayer._getinfo.ClientSession.post",
    side_effect=mock_session_post,
)
async def test_get_json_dict(mock_response):
    async with ClientSession() as s:
        assert await get_jsondict("test", session=s) == TESTJSON


@pytest.mark.asyncio
@patch(
    "restgdf.featurelayer._getinfo.ClientSession.post",
    side_effect=mock_session_post,
)
async def test_get_feature_count(mock_response):
    async with ClientSession() as s:
        assert await get_feature_count("test", session=s) == TESTJSON["count"]
        assert await get_feature_count("test", session=s, data={"where": "test"})
        assert await get_feature_count("test", session=s, data={"token": "test"})


@patch(
    "restgdf.featurelayer._getinfo.ClientSession.post",
    side_effect=mock_session_post,
)
def test_get_max_record_count(mock_response):
    assert get_max_record_count(TESTJSON) == TESTJSON["maxRecordCount"]


@pytest.mark.asyncio
@patch(
    "restgdf.featurelayer._getinfo.ClientSession.post",
    side_effect=mock_session_post,
)
async def test_get_offset_range(mock_response):
    async with ClientSession() as s:
        assert await get_offset_range("test", session=s) == range(
            0,
            TESTJSON["count"],
            TESTJSON["maxRecordCount"],
        )


def test_default_data():
    t1 = {}
    t1r = {k: v for k, v in DEFAULTDICT.items()}

    t2 = {"where": "test"}
    t2r = {k: v for k, v in DEFAULTDICT.items()}
    t2r["where"] = "test"

    t3 = {"token": "test"}
    t3r = {k: v for k, v in DEFAULTDICT.items()}
    t3r["token"] = "test"

    for indict, outdict in zip([t1, t2, t3], [t1r, t2r, t3r]):
        assert default_data(indict) == outdict
